import http from 'http';
import {BackendState} from './common';
import {Language, Resource, Charset, MediaType, LanguageStatusMessageMap} from '../common';
import https from 'https';
import Negotiator from 'negotiator';
import {constants} from 'http2';
import {
	handleCreateItem,
	handleDeleteItem,
	handleEmplaceItem,
	handleGetCollection,
	handleGetItem,
	handleGetRoot,
	handlePatchItem,
} from './handlers';
import {
	BackendResource,
} from './core';
import * as v from 'valibot';
import {getBody} from './utils';
import {DataSource} from './data-source';

export interface Response {
	statusCode: number;

	statusMessage?: keyof LanguageStatusMessageMap;

	headers?: Record<string, string>;
}

interface ResponseParams {
	statusCode: Response['statusCode'];
	statusMessage?: Response['statusMessage'];
	headers?: Response['headers'];
}

export class MiddlewareError extends Error {}

interface PlainResponseParams<T = unknown> extends ResponseParams {
	body?: T;
}

interface HttpMiddlewareErrorParams<T = unknown> extends Omit<PlainResponseParams<T>, 'statusMessage'> {
	cause?: unknown
}

export class PlainResponse<T = unknown> implements Response {
	readonly statusCode: Response['statusCode'];

	readonly statusMessage?: keyof LanguageStatusMessageMap;

	readonly headers: Response['headers'];

	readonly body?: T;

	constructor(args: PlainResponseParams<T>) {
		this.statusCode = args.statusCode;
		this.statusMessage = args.statusMessage;
		this.headers = args.headers;
		this.body = args.body;
	}
}

export class HttpMiddlewareError extends MiddlewareError {
	readonly response: PlainResponse;

	constructor(statusMessage: keyof Language['statusMessages'], params: HttpMiddlewareErrorParams) {
		super(statusMessage, { cause: params.cause });
		this.response = new PlainResponse({
			...params,
			statusMessage,
		});
	}
}

export interface CreateServerParams {
	basePath?: string;
	host?: string;
	cert?: string;
	key?: string;
	requestTimeout?: number;
	// CQRS
	streamResponses?: boolean;
}

export interface RequestContext extends http.IncomingMessage {
	backend?: BackendState;

	host?: string;

	scheme?: string;

	basePath?: string;

	method?: string;

	url?: string;

	rawUrl?: string;

	cn: {
		language: Language;
		mediaType: MediaType;
		charset: Charset;
	};

	query: URLSearchParams;

	resource: BackendResource;

	resourceId?: string;

	body?: unknown;
}

export interface Middleware<Req extends RequestContext = RequestContext> {
	(req: Req): undefined | Response | Promise<undefined | Response>;
}

const getAllowedMiddlewares = <T extends v.BaseSchema>(resource?: Resource<T>, mainResourceId = '') => {
	const middlewares = [] as [string, Middleware, v.BaseSchema?][];

	if (typeof resource === 'undefined') {
		return middlewares;
	}

	if (mainResourceId === '') {
		if (resource.state.canFetchCollection) {
			middlewares.push(['GET', handleGetCollection]);
		}
		if (resource.state.canCreate) {
			middlewares.push(['POST', handleCreateItem, resource.schema]);
		}
		return middlewares;
	}

	if (resource.state.canFetchItem) {
		middlewares.push(['GET', handleGetItem]);
	}
	if (resource.state.canEmplace) {
		const schema = resource.schema.type === 'object' ? resource.schema as unknown as v.ObjectSchema<any> : resource.schema;
		const putSchema = (
			schema.type === 'object'
				? v.merge([
					schema as v.ObjectSchema<any>,
					v.object({
						[resource.state.idAttr]: v.transform(
							v.any(),
							input => resource.state.idConfig.serialize(input),
							v.literal(mainResourceId)
						)
					})
				])
				: schema
		);
		middlewares.push(['PUT', handleEmplaceItem, putSchema]);
	}
	if (resource.state.canPatch) {
		const schema = resource.schema.type === 'object' ? resource.schema as unknown as v.ObjectSchema<any> : resource.schema;
		const patchSchema = (
			schema.type === 'object'
				? v.partial(
					schema as v.ObjectSchema<any>,
					(schema as v.ObjectSchema<any>).rest,
					(schema as v.ObjectSchema<any>).pipe
				)
				: schema
		);
		middlewares.push(['PATCH', handlePatchItem, patchSchema]);
	}
	if (resource.state.canDelete) {
		middlewares.push(['DELETE', handleDeleteItem]);
	}

	return middlewares;
};

const adjustRequestForContentNegotiation = (req: RequestContext, res: http.ServerResponse<RequestContext>) => {
	const negotiator = new Negotiator(req);
	const availableLanguages = Array.from(req.backend!.app.languages);
	const availableCharsets = Array.from(req.backend!.app.charsets);
	const availableMediaTypes = Array.from(req.backend!.app.mediaTypes);

	const languageCandidate = negotiator.language(availableLanguages.map((l) => l.name)) ?? req.backend!.cn.language.name;
	const charsetCandidate = negotiator.charset(availableCharsets.map((l) => l.name)) ?? req.backend!.cn.charset.name;
	const mediaTypeCandidate = negotiator.mediaType(availableMediaTypes.map((l) => l.name)) ?? req.backend!.cn.mediaType.name;

	// TODO refactor
	const currentLanguage = availableLanguages.find((l) => l.name === languageCandidate);
	if (typeof currentLanguage === 'undefined') {
		const data = req.backend?.cn.language.bodies.languageNotAcceptable;
		const responseRaw = req.backend?.cn.mediaType.serialize(data);
		const response = typeof responseRaw !== 'undefined' ? req.backend?.cn.charset.encode(responseRaw) : undefined;
		res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
			'Content-Language': req.backend?.cn.language.name,
			'Content-Type': [
				req.backend?.cn.mediaType.name,
				`charset="${req.backend?.cn.charset.name}"`
			].join('; '),
		});
		res.statusMessage = req.backend?.cn.language.statusMessages.languageNotAcceptable ?? '';
		res.end(response);
		return;
	}

	const currentMediaType = availableMediaTypes.find((l) => l.name === mediaTypeCandidate);
	if (typeof currentMediaType === 'undefined') {
		const data = req.backend?.cn.language.bodies.mediaTypeNotAcceptable;
		const responseRaw = req.backend?.cn.mediaType.serialize(data);
		const response = typeof responseRaw !== 'undefined' ? req.backend?.cn.charset.encode(responseRaw) : undefined;
		res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
			'Content-Language': req.backend?.cn.language.name,
			'Content-Type': [
				req.backend?.cn.mediaType.name,
				`charset="${req.backend?.cn.charset.name}"`
			].join('; '),
		});
		res.statusMessage = req.backend?.cn.language.statusMessages.mediaTypeNotAcceptable ?? '';
		res.end(response);
		return;
	}

	const responseBodyCharset = availableCharsets.find((l) => l.name === charsetCandidate);
	if (typeof responseBodyCharset === 'undefined') {
		const data = req.backend!.cn.language.bodies.encodingNotAcceptable;
		const responseRaw = req.backend!.cn.mediaType.serialize(data);
		const response = typeof responseRaw !== 'undefined' ? req.backend!.cn.charset.encode(responseRaw) : undefined;
		res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
			'Content-Language': req.backend?.cn.language.name,
			'Content-Type': [
				req.backend?.cn.mediaType.name,
				`charset="${req.backend?.cn.charset.name}"`
			].join('; '),
		});
		res.statusMessage = req.backend?.cn.language.statusMessages.encodingNotAcceptable ?? '';
		res.end(response);
		return;
	}

	req.cn.language = currentLanguage;
	req.cn.mediaType = currentMediaType;
	req.cn.charset = responseBodyCharset;
};

export const createServer = (backendState: BackendState, serverParams = {} as CreateServerParams) => {
	const isHttps = 'key' in serverParams && 'cert' in serverParams;

	const server = isHttps
		? https.createServer({
			key: serverParams.key,
			cert: serverParams.cert,
			requestTimeout: serverParams.requestTimeout,
		})
		: http.createServer({
			requestTimeout: serverParams.requestTimeout,
		});

	server.on('request', async (req: RequestContext, res: http.ServerResponse<RequestContext>) => {
		req.backend = backendState;
		req.basePath = serverParams.basePath ?? '';
		req.host = serverParams.host ?? 'localhost';
		req.scheme = isHttps ? 'https' : 'http';
		req.cn = req.backend.cn;
		req.method = req.method?.trim().toUpperCase() ?? '';

		const theBasePathUrl = req.basePath ?? '';
		const basePath = new URL(theBasePathUrl, 'http://localhost');
		const parsedUrl = new URL(`${theBasePathUrl}/${req.url ?? ''}`, 'http://localhost');
		req.rawUrl = req.url;
		req.url = req.url?.slice(basePath.pathname.length) ?? '';
		req.query = parsedUrl.searchParams;

		adjustRequestForContentNegotiation(req, res);

		let middlewareState;
		if (req.url === '/' || req.url === '') {
			middlewareState = await handleGetRoot(req);
		}

		if (typeof middlewareState === 'undefined') {
			const [, resourceRouteName, resourceId = ''] = req.url?.split('/') ?? [];
			const resource = Array.from(req.backend.app.resources).find((r) => r.state!.routeName === resourceRouteName);
			if (typeof resource === 'undefined') {
				res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
				res.statusMessage = req.backend.cn.language.statusMessages.urlNotFound;
				res.end();
				return;
			}

			req.resource = resource as BackendResource;
			req.resource.dataSource = req.backend.dataSource(req.resource) as DataSource;
			req.resourceId = resourceId;

			try {
				await req.resource.dataSource.initialize();
			} catch (cause) {
				throw new HttpMiddlewareError(
					'unableToInitializeResourceDataSource',
					{
						cause,
						statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
					}
				);
			}
		}

		const middlewares = getAllowedMiddlewares(req.resource, req.resourceId ?? '');
		middlewareState = await middlewares.reduce<unknown>(
			async (currentHandlerStatePromise, currentValue) => {
				const [middlewareMethod, middleware, schema] = currentValue;
				try {
					const currentHandlerState = await currentHandlerStatePromise;
					const effectiveMethod = req.method === 'HEAD' ? 'GET' : req.method;

					if (effectiveMethod !== middlewareMethod) {
						return currentHandlerState;
					}

					if (typeof currentHandlerState !== 'undefined') {
						return currentHandlerState;
					}

					if (schema) {
						const availableSerializers = Array.from(req.backend!.app.mediaTypes);
						const availableCharsets = Array.from(req.backend!.app.charsets);
						const contentTypeHeader = req.headers['content-type'] ?? 'application/octet-stream';
						const fragments = contentTypeHeader.split(';');
						const mediaType = fragments[0];
						const charsetParam = fragments.map((s) => s.trim())
							.find((f) => f.startsWith('charset=')) ?? (mediaType.startsWith('text/') ? 'charset=utf-8' : 'charset=binary');
						const [_charsetKey, charsetRaw] = charsetParam.split('=').map((s) => s.trim());
						const charset = (
							(
								(charsetRaw.startsWith('"') && charsetRaw.endsWith('"'))
								|| (charsetRaw.startsWith("'") && charsetRaw.endsWith("'"))
							)
								? charsetRaw.slice(1, -1).trim()
								: charsetRaw.trim()
						)
						const deserializerPair = availableSerializers.find((l) => l.name === mediaType);
						const encodingPair = availableCharsets.find((l) => l.name === charset);
						(req as unknown as Record<string, unknown>).body = await getBody(req, schema, encodingPair, deserializerPair);
					}

					const result = await middleware(req);

					if (req.method === 'HEAD' && result instanceof PlainResponse) {
						const { body: _, ...etcResult } = result;

						return Promise.resolve(new PlainResponse(etcResult));
					}

					return Promise.resolve(result);
				} catch (errRaw) {
					// todo use error message key for each method
					// TODO better error reporting, localizable messages
					// TODO handle error handlers' errors
					if (errRaw instanceof v.ValiError && Array.isArray(errRaw.issues)) {
						return new HttpMiddlewareError('invalidResource', {
							statusCode: constants.HTTP_STATUS_BAD_REQUEST,
							body: errRaw.issues.map((i) => (
								`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
							)),
						});
					}

					return errRaw;
				}
			},
			Promise.resolve<ReturnType<Middleware> | HttpMiddlewareError>(middlewareState)
		) as Awaited<ReturnType<Middleware> | HttpMiddlewareError>;

		if (typeof middlewareState !== 'undefined') {
			try {
				if (middlewareState instanceof Error) {
					throw middlewareState;
				}

				const headers: Record<string, string> = {
					...(
						middlewareState.headers ?? {}
					),
					'Content-Language': req.cn.language.name
				};

				if (middlewareState instanceof http.ServerResponse) {
					// TODO streaming responses
					middlewareState.writeHead(constants.HTTP_STATUS_ACCEPTED, headers);
					return;
				}

				if (middlewareState instanceof PlainResponse) {
					let encoded: Buffer | undefined;
					if (typeof middlewareState.body !== 'undefined') {
						let serialized;
						try {
							serialized = req.cn.mediaType.serialize(middlewareState.body);
						} catch (cause) {
							throw new HttpMiddlewareError('unableToSerializeResponse', {
								cause,
								statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
								headers: {
									'Content-Language': req.backend.cn.language.name,
								},
							})
						}

						try {
							encoded = req.cn.charset.encode(serialized);
						} catch (cause) {
							throw new HttpMiddlewareError('unableToEncodeResponse', {
								cause,
								statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
								headers: {
									'Content-Language': req.backend.cn.language.name,
								},
							})
						}

						headers['Content-Type'] = [
							req.cn.mediaType.name,
							`charset=${req.cn.charset.name}`
						].join('; ');
					}

					const statusMessageKey = middlewareState.statusMessage ? req.cn.language.statusMessages[middlewareState.statusMessage] : undefined;
					res.statusMessage = statusMessageKey?.replace(/\$RESOURCE/g, req.resource.state.itemName) ?? '';
					res.writeHead(middlewareState.statusCode, headers);
					if (typeof encoded !== 'undefined') {
						res.end(encoded);
						return;
					}
					res.end();
				}
				return;
			} catch (finalErrRaw) {
				const finalErr = finalErrRaw as HttpMiddlewareError;
				const headers = finalErr.response.headers ?? {};
				let encoded: Buffer | undefined;
				let serialized;
				try {
					serialized = typeof finalErr.response.body !== 'undefined' ? req.backend.cn.mediaType.serialize(finalErr.response.body) : undefined;
				} catch (cause) {
					res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR);
					res.end();
					return;
				}

				try {
					encoded = typeof serialized !== 'undefined' ? req.backend.cn.charset.encode(serialized) : undefined;
				} catch (cause) {
					res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR);
					res.end();
				}

				headers['Content-Type'] = [
					req.backend.cn.mediaType.name,
					`charset=${req.backend.cn.charset.name}`
				].join('; ');

				const statusMessageKey = finalErr.response.statusMessage ? req.backend.cn.language.statusMessages[finalErr.response.statusMessage] : undefined;
				res.statusMessage = statusMessageKey?.replace(/\$RESOURCE/g, req.resource.state.itemName) ?? '';
				res.writeHead(finalErr.response.statusCode, headers);
				if (typeof encoded !== 'undefined') {
					res.end(encoded);
					return;
				}
				res.end();
				return;
			}
		}

		if (middlewares.length > 0) {
			res.statusMessage = req.backend.cn.language.statusMessages.methodNotAllowed.replace(/\$RESOURCE/g, req.resource.state.itemName) ?? '';
			res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
				Allow: middlewares.map((m) => m[0]).join(', '),
				'Content-Language': req.backend.cn.language.name,
			});
			res.end();
			return;
		}

		// TODO error handler in line with authentication
		res.statusMessage = req.backend.cn.language.statusMessages.urlNotFound.replace(/\$RESOURCE/g, req.resource.state.itemName) ?? '';
		res.writeHead(constants.HTTP_STATUS_NOT_FOUND, {
			'Content-Language': req.backend.cn.language.name,
		});
		res.end();
		return;
	});

	return server;
}
